/**************************************************************************
*
*	Controller.h
*
*	Controller API class, which makes it possible to control robots!
*
*	Author: Saxon Druce
*
*	Copyright (c) 2004 Block Software Pty Ltd
*
**************************************************************************/

#ifndef _CONTROLLER_H
#define _CONTROLLER_H

/*************************************************************************/
// Include files
/*************************************************************************/

#include <vector>

#include "CommsClient.h"
#include "Vec.h"

/*************************************************************************/
// Defines
/*************************************************************************/

/*************************************************************************/
// Public types
/*************************************************************************/

// Controller_t class - the class used to control a robot
class Controller_t : private CommsClient_t
{
public:

	// SensedRobot_t class - holds details of a robot which has been sensed
	struct SensedRobot_t
	{
	public:

		// the robot's ID
		int m_RobotID;

		// the robot's position
		Vec_t m_Position;

		// the robot's orientation
		double m_Orientation;

		// the robot's name
		std::string m_Name;

		// the robot's owner
		std::string m_Owner;
	};

	// SensedMunition_t class - holds details of a munition which has been sensed
	struct SensedMunition_t
	{
	public:

		// the munition's ID
		int m_MunitionID;

		// the munition's position
		Vec_t m_Position;
	};

	// Constructor
	Controller_t();

	// Destructor
	~Controller_t();

	// Connect() - Connects to the specified port on the specified host
	// (name or IP address). Throws an exception on failure to connect.
	// void Connect(
	//	const std::string& HostName,
	//	int Port);
	using CommsClient_t::Connect; // pull into public use from the base class

	// GetPlayingFieldSize() - Gets the size (width and height) of the 
	// playing field, in metres. The origin (0, 0) is in the middle of 
	// the playing field, so the field extends from -1/2 to +1/2 of the 
	// width and height in the x and y directions respectively.
	void GetPlayingFieldSize(double& Width, double& Height);

	// CreateRobot() - Creates a robot, returning its ID for reference.
	// A controller can only create a single robot at any one time.
	int CreateRobot();

	// NOTE: All of the following API methods apply to the robot which was
	// created with the last call to CreateRobot(). They will fail if a
	// robot has not been created.

	// RemoveRobot() - Removes the robot from the simulation. If you
	// disconnect while a robot you have created exists, it will be
	// automatically removed.
	void RemoveRobot();

	// SetName() - Sets the robot's name.
	void SetName(const std::string& Name);

	// SetOwner() - Sets the robot's owner.
	void SetOwner(const std::string& Owner);

	// SetColour() - Sets the colour which viewers will draw the robot
	// in. Each component is a value from 0 to 255, although the sum of
	// the thee components must be less than 450 (otherwise the colour
	// will be too light to see on a white background). The robot will be
	// drawn with a default colour if one is not provided.
	void SetColour(int Red, int Green, int Blue);

	// Taunt() - Sends a message to the server which will be sent to all 
	// viewers.
	void Taunt(const std::string& Message);

	// GetNumKills() - Returns the number of robots which the robot has 
	// killed.
	int GetNumKills();

	// GetKiller() - If this robot has been killed (health is zero), returns 
	// the ID, name and owner of the robot which killed it.
	void GetKiller(int& RobotID, std::string& Name, std::string& Owner);

	// GetMoney() - Returns the amount of money which the robot has
	// remaining for purchasing equipment.
	int GetMoney();

	// PurchaseEngine() - Purchases an engine with the specified parameters 
	// for the robot.
	void PurchaseEngine(double TopSpeed, double TurningSpeed);

	// PurchaseSensor() - Purchases a sensor with the specified parameters
	// for the robot.
	void PurchaseSensor(
		double Range, double BeamWidth, double TurningSpeed);

	// PurchaseBattery() - Purchases a battery with the specified parameters
	// for the robot.
	void PurchaseBattery(double Capacity);

	// PurchaseTurret() - Purchases a turret with the specified parameters 
	// for the robot.
	void PurchaseTurret(double MovementRange, double TurningSpeed);

	// PurchaseWeapon() - Purchases a weapon with the specified parameters
	// for the robot.
	void PurchaseWeapon(
		double ChargeTime, double EnergyUsage, double Damage, double MunitionSpeed);

	// PurchaseSolarPanel() - Purchases a solar panel with the specified 
	// parameters the robot.
	void PurchaseSolarPanel(double Efficiency);

	// EnterSimulation() - Enters the robot into active simulation, by
	// placing it at a random location and orientation on the playing
	// field.
	void EnterSimulation();

	// NOTE: The following methods can only be called once the robot has
	// entered the simulation.

	// GetHealth() - Returns the current health of the robot, a value from
	// 0 (dead) to 100 (full).
	double GetHealth();

	// NOTE: The following methods can only be called if the robot is
	// running, ie has entered the simulation and hasn't been killed.

	// GetPosition() - Returns the current position of the robot.
	Vec_t GetPosition();

	// GetOrientation() - Returns the current orientation of the robot,
	// ie the angle with the x-axis in radians, positive clockwise.
	double GetOrientation();

	// SetThrottle() - Sets the robot's throttle value, from -100 
	// (backwards at 100% of engine's top speed) to 100 (forwards at
	// 100% of engine's top speed).
	void SetThrottle(double Throttle);

	// SetDesiredOrientation() - Sets the robot's desired orientation,
	// ie the angle with the x-axis in radians, positive clockwise.
	// Once the desired orientation is set, the robot will turn as
	// fast as its engine allows from its current orientation to the
	// desired orientation.
	void SetDesiredOrientation(double DesiredOrientation);

	// GetSensorOrientation() - Returns the current orientation of the
	// sensor, in radians, where 0 means pointing in the same direction
	// as the robot, positive is clockwise (to the right).
	double GetSensorOrientation();

	// SetDesiredSensorOrientation() - Sets the robot's desired sensor
	// orientation, ie the angle relative to the robot's orientation,
	// positive clockwise. Once the desired orientation is set, the 
	// sensor will turn as fast as possible from its current orientation 
	// to the desired orientation.
	void SetDesiredSensorOrientation(
		double DesiredSensorOrientation);

	// GetSensedRobots() - Returns a list of all robots which the sensor
	// of the robot can currently see.
	std::vector<SensedRobot_t> GetSensedRobots();

	// GetSensedMunitions() - Returns a list of all munitions which the 
	// sensor of the robot can currently see.
	std::vector<SensedMunition_t> GetSensedMunitions();

	// GetEnergy() - Returns the robot's current energy level.
	double GetEnergy();

	// GetTurretOrientation() - Returns the current orientation of the
	// turret, in radians, where 0 means pointing in the same direction
	// as the robot, positive is clockwise (to the right).
	double GetTurretOrientation();

	// SetDesiredTurretOrientation() - Sets the robot's desired turret
	// orientation, ie the angle relative to the robot's orientation,
	// positive clockwise. Once the desired orientation is set, the 
	// turret will turn as fast as possible from its current orientation 
	// to the desired orientation.
	void SetDesiredTurretOrientation(
		double DesiredTurretOrientation);

	// GetWeaponCharge() - Returns the percentage that the weapon is
	// charged, from 0 to 100.
	double GetWeaponCharge();

	// FireWeapon() - Fires the robot's weapon.
	void FireWeapon();
};

/*************************************************************************/

#endif // _CONTROLLER_H
