/**************************************************************************
*
*	SimpleRobot.cs
*
*	Main source file for simple robot sample programme (in C#)
*
*	Author: Saxon Druce
*
*	Copyright (c) 2004 Block Software Pty Ltd
*
**************************************************************************/

using System;
using System.Collections;
using ControllerNet;
using VecNet;
using GraphicsNet;

namespace SimpleRobotNet
{
	/// <summary>
	/// The main class.
	/// </summary>
	class SimpleRobot
	{
		/// <summary>
		/// random number generator
		/// </summary>
		static Random m_random = new Random();

		// IDs to use for debug graphics
		static int GRAPHIC_ID_PLAYING_FIELD	= 1;
		static int GRAPHIC_ID_ROBOT	= 2;
		static int GRAPHIC_ID_SEARCH_TARGET = 3;
		static int GRAPHIC_ID_ENEMY = 4;

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			try
			{
				// setup default values of parameters
				string serverName = "localhost";
				bool debugGraphics = false;

				// expect arguments to be in the following format:
				// /a a_value /b b_value
				// so there should be an even number of them
				if (args.Length % 2 != 0)
					throw new Exception("Expected an even number of arguments");

				// now loop over them
				int i;
				for (i = 0; i < args.Length; i += 2)
				{
					if (args[i] == "/s")
					{
						// /s = server address (a string)
						serverName = args[i + 1];
					}
					else if (args[i] == "/d")
					{
						// /d = use debug graphics (an int, 1 = yes, 0 = no)
						debugGraphics = (Int32.Parse(args[i + 1]) != 0);
					}
					else
					{
						throw new Exception("Invalid argument " + args[i]);
					}
				}

				// create a controller
				Controller controller = new Controller();

				// connect
				controller.Connect(serverName, 9386);

				// create a graphics client
				Graphics graphics = new Graphics();

				// if debugging, connect (on localhost)
				if (debugGraphics)
					graphics.Connect("localhost", 9387);

				// get the playing field size
				double playingFieldWidth, playingFieldHeight;
				controller.GetPlayingFieldSize(
					out playingFieldWidth, out playingFieldHeight);
				Console.WriteLine("Playing field size is {0}m x {1}m", 
					playingFieldWidth, playingFieldHeight);

				// if debugging, add the playing field for reference
				if (debugGraphics)
				{
					graphics.SetToRectangle(
						GRAPHIC_ID_PLAYING_FIELD,
						new Vec(0, 0),				// centre position
						playingFieldWidth,
						playingFieldHeight,
						0,							// orientation
						4,							// line width
						0, 0, 139);					// colour
				}

				// create a robot
				int robotID = controller.CreateRobot();
				Console.WriteLine("Robot ID is {0}", robotID);

				// set name and owner
				controller.SetName("Simple Robot C#");
				controller.SetOwner("Saxon");

				// set colour
				controller.SetColour(0, 0, 255);

				// get initial amount of money
				Console.WriteLine("Robot has ${0} to spend on equipment", 
					controller.GetMoney());

				// buy equipment
				controller.PurchaseEngine(2, 50);
				controller.PurchaseSensor(10, 60, 40);
				controller.PurchaseBattery(50);
				controller.PurchaseTurret(0, 0);
				controller.PurchaseWeapon(2, 4, 2, 3);
				controller.PurchaseSolarPanel(1);

				// get amount of money left
				Console.WriteLine("Robot has ${0} remaining after purchasing equipment", 
					controller.GetMoney());

				// enter the simulation
				controller.EnterSimulation();

				// set up a value to hold the position which the robot is
				// trying to get to, and initialise it with the robot's
				// current position
				Vec targetPosition = controller.GetPosition();

				// begin assuming that we haven't killed anybody yet
				int lastNumKills = 0;

				try
				{
					// loop forever, since there doesn't seem to be a way
					// to loop until a key is pressed
					Console.WriteLine("\nUse ctrl+C to stop\n");
					for (;;)
					{
						// get the robot's current position
						Vec currentPosition = controller.GetPosition();

						// get the robot's current orientation
						double currentOrientation = controller.GetOrientation();

						// if debugging, draw the robot
						if (debugGraphics)
						{
							graphics.SetToRectangle(
								GRAPHIC_ID_ROBOT,
								currentPosition,
								1.0,					// length
								0.6,					// 'height' (width of robot)
								currentOrientation,
								1,						// line width
								0, 0, 255);				// colour
						}

						// sense other robots
						ArrayList robots = controller.GetSensedRobots();

						if (robots.Count == 0)
						{
							// we couldn't see any robots, so simply wander randomly
							// around the playing field, in the hope that we'll
							// eventually find another robot

							// work out the vector from the current position to the
							// target position
							Vec targetVector = targetPosition - currentPosition;

							if (targetVector.Length() < 1.0)
							{
								// we're within 1 metre of the target, so assume we've
								// reached it - therefore pick a new random target
								// position, which is at least 2 metres from the edge
								// of the playing field
								targetPosition.x = RandomBetween(
									-(playingFieldWidth / 2.0 - 2.0), 
									(playingFieldWidth / 2.0 - 2.0));
								targetPosition.y = RandomBetween(
									-(playingFieldHeight / 2.0 - 2.0), 
									(playingFieldHeight / 2.0 - 2.0));

								Console.WriteLine(
									"New target position is {0:f1}, {0:f1}",
									targetPosition.x,
									targetPosition.y);
						
								// recalculate the target vector since it is used below
								targetVector = targetPosition - currentPosition;
							}

							// work out the desired orientation to point towards the
							// target
							double desiredOrientation = targetVector.GetAngle();

							// set it
							controller.SetDesiredOrientation(
								desiredOrientation);

							// work out the difference between the current orientation and
							// the desired orientation
							double orientationDifference = 
								desiredOrientation - currentOrientation;
					
							// wrap to -pi to pi range
							if (orientationDifference < -Math.PI)
								orientationDifference += 2 * Math.PI;
							else if (orientationDifference >= Math.PI)
								orientationDifference -= 2 * Math.PI;

							if (Math.Abs(orientationDifference) < 5 * Math.PI / 180)
							{
								// we're within 5 degrees of the desired orientation, so
								// set the throttle to full forwards
								controller.SetThrottle(100.0);
							}
							else
							{
								// set the throttle to stay still while we turn
								controller.SetThrottle(0.0);
							}

							// set sensor orientation to 90 degrees further clockwise
							// than it currently is, which should result in a constant
							// clockwise spin as fast as possible (assuming we aren't
							// able to spin by 90 degrees between iterations of this loop)
							controller.SetDesiredSensorOrientation(
								controller.GetSensorOrientation() + Math.PI / 2);

							// if debugging, draw a circle at the point we're targeting,
							// and hide any indication of an enemy
							if (debugGraphics)
							{
								graphics.SetToCircle(
									GRAPHIC_ID_SEARCH_TARGET,
									targetPosition,
									1.0,				// radius
									1,					// line width
									50, 200, 50);		// colour

								graphics.Remove(GRAPHIC_ID_ENEMY);
							}
						}
						else
						{
							// we can see at least one robot, so attack the first one

							// get it
							SensedRobot enemy = (robots[0] as SensedRobot);

							// work out the vector from the current position to the
							// enemy robot position
							Vec enemyVector = 
								enemy.m_Position - currentPosition;

							// work out the desired orientation to point towards the
							// enemy robot
							double desiredOrientation = enemyVector.GetAngle();

							// set it
							controller.SetDesiredOrientation(
								desiredOrientation);

							// work out the difference between the current orientation and
							// the desired orientation
							double orientationDifference = 
								desiredOrientation - currentOrientation;
					
							// wrap to -pi to pi range
							if (orientationDifference < -Math.PI)
								orientationDifference += 2 * Math.PI;
							else if (orientationDifference >= Math.PI)
								orientationDifference -= 2 * Math.PI;

							if (Math.Abs(orientationDifference) < 5 * Math.PI / 180)
							{
								// we're within 5 degrees of the desired orientation

								// work out how far we are from the enemy
								double enemyDistance = enemyVector.Length();

								if (enemyDistance < 5)
								{
									// within 5 metres of the enemy, so move
									// backwards to keep away from it (we'll
									// be too easy to hit if we get too close)
									controller.SetThrottle(-100.0);
								}
								else if (enemyDistance < 6)
								{
									// between 5 and 6 metres from the enemy,
									// so stay still
									controller.SetThrottle(0.0);
								}
								else
								{
									// more than 6 metres from the enemy, so
									// move towards it
									controller.SetThrottle(100.0);
								}

								if (enemyDistance < 6)
								{
									// we're within a reasonable range of the enemy,
									// so fire the weapon if it's fully charged
									if (controller.GetWeaponCharge() >= 100)
										controller.FireWeapon();
								}
							}
							else
							{
								// set the throttle to stay still while we turn
								controller.SetThrottle(0.0);
							}

							// set the sensor angle required to keep it pointing 
							// at the enemy, to make sure it doesn't get away
							controller.SetDesiredSensorOrientation(
								desiredOrientation - currentOrientation);

							// if debugging, draw a circle at the point of the enemy,
							// and hide any search target
							if (debugGraphics)
							{
								graphics.SetToCircle(
									GRAPHIC_ID_ENEMY,
									enemy.m_Position,
									0.5,				// radius
									1,					// line width
									255, 0, 0);			// colour

								graphics.Remove(GRAPHIC_ID_SEARCH_TARGET);
							}
						}

						// check if we've killed anybody else
						int numKills = controller.GetNumKills();
						if (lastNumKills != numKills)
						{
							controller.Taunt("Woohoo I killed somebody!");
							lastNumKills = numKills;
						}
					}
				}
				catch (Exception e)
				{
					// check if the exception was because we tried to
					// control the robot after it has been killed
					if (controller.GetHealth() <= 0)
					{
						// we've been killed - get details of our killer
						int killerID;
						string killerName, killerOwner;
						controller.GetKiller(out killerID, out killerName, out killerOwner);

						// then send a parting message
						controller.Taunt(
							"I don't like you " + 
							killerName + " / " + killerOwner + " !!");
					}
					else
					{
						// the exception was for some other reason, so propagate it
						throw e;
					}
				}
			}
			catch (Exception e)
			{
				Console.WriteLine(
					"An exception occurred: {0}",
					e.Message);
			}
		}

		/// <summary>
		/// Helper function to return a random value between the two bounds.
		/// </summary>
		static double RandomBetween(double lowerBound, double upperBound)
		{
			return lowerBound + m_random.NextDouble() * (upperBound - lowerBound);
		}
	}
}
